import math
import openscad
from pyopenscad import * # don't do this
from copy import copy
from door_util import calc_door_offsets
from vector_util import * # yeah don't do this

# heights
h = 1.2
h0 = 0.0
h1 = 0.55
h2 = h1 + 0.22
h3 = 1.05
hpoints = [h0, h1, h2, h3, h]

# widths (half-widths)
w = 0.75
w0 = 0
w1 = 0.2
w2 = 0.3
w3 = w-0.08
wpoints = [w0, w1, w2, w3, w]

spline_me = True

# front portion depths
front_fwd = []
d = 0.0
for p in [0.0, 0.29, 0.47, 0.34]:
    d += p
    front_fwd.append(d)

# middle portion depths
middle_fwd = []
d = 0.0
for p in [0.0, 0.2, 0.3, 0.3, 0.3]:
    d += p
    middle_fwd.append(d)

# back forward depths (rear wheel)
back_fwd = []
d = 0.0
for p in [0.0, 0.23, 0.61, 0.34]:
    d += p
    back_fwd.append(d)

# wheel radius: diameter inches plus tire, convert to cm, divide by 2
# calculator.net/tire-size-calculator.html 165/70R13
wheel_radius = (22.1)*2.54/2. / 100.0
wheel_inner = (13)*2.54/2. / 100.0
wtire = 6.5*2.54/100.0 # 6.5in tire tread

wheel_angle = 00.0 # rotate front wheels

# move the depth of the cockpit floor inwards at the backbox point
# gives a bit more room to put the rear wheel suspension arm and motor in
cockpit_floor_slope = 0.05

# pipe diameters
dthin = 0.013
dthick = 0.0215

def openscad_save(o, name, as_stl=False):
    scadpth = os.path.join(".", "scad")
    stlpth = os.path.join(".", "scad", "stl")
    try:
        os.makedirs(scadpth)
    except OSError:
        pass
    try:
        os.makedirs(stlpth)
    except OSError:
        pass

    scad_name = os.path.join(scadpth, name + ".scad")
    stl_name = os.path.join(stlpth, name + ".stl")

    scad_render_to_file(o, scad_name)
    if as_stl:
        print "generating stl for ", name
        openscad.run("-o", stl_name, scad_name)


def openscad_to_stl(o, name):
    openscad_save(o, name, as_stl=True)


def pipe(l, radius=0.025 / 2.0, thick = 0.0025, offs=0.05, # 2.5mm thickness
            cutangle1=None, cutangle2=None, radius2=None):
    u = difference()
    height = l-offs
    #u = union()
    u.add(cylinder(r=radius, r1=radius2, h=height, segments=20, center=True))
    # XXX massively slows down
    #u.add(cylinder(r=radius-thick, h=l-offs+0.001, segments=20, center=True))
    cut = union()
    c = cube([radius*1.5, radius*2.2, radius*0.7])
    c = rotate([0,45,0]).add(c)
    r = translate([0,-radius*1.101,0]).add(c)
    cut.add(r)
    r = rotate([0,0,180]).add(r)
    cut.add(r)
    cut = rotate([0,0,45]).add(cut)
    if cutangle1 is not None:
        r = translate([0,0,height/2]).add(cut)
        r = rotate([0,0,cutangle1]).add(r)
        u.add(r)
    # cheat and drop length into the result
    u.length = height
    return u

def ppipe(p1, p2, radius=0.025, offs=0.05, cutangle1=None, cutangle2=None,
          radius2=None):
    l = pdist(p1, p2)
    return pipe(l, radius, offs=offs, cutangle1=cutangle1, cutangle2=cutangle2,
                   radius2=radius2)

def boxsect(length, width, height, center=False, thick=0.003):
    u = difference()
    c = cube([width, length, height], center=center)
    u.add(c)
    c = cube([width-thick*2, length+thick*2, height-thick*2], center=center)
    if center:
        c = translate([thick/2.0, -thick/2.0, thick/2.0]).add(c)
    else:
        c = translate([thick, -thick, thick]).add(c)
    u.add(c)
    return u


def box50100(length, center=False, thick=0.003):
    return boxsect(length, 0.1, 0.05, center=center, thick=thick)


# joint up/down offsets
d1 = 1.0
d2 = pow(d1*2, 0.5) / 2.0
d3 = pow(d2*2, 0.5) / 2.0
jup = [0,0,d1]
jdown = [0,0,-d1]
jfwd = [0,-d1,0]
jback = [0,d1,0]
jleft = [-d1,0,0]
jright = [d1,0,0]
jrf = [d2,-d2,0]
jlf = [-d2,-d2,0]
jrb = [d2,d2,0]
jlb = [-d2,d2,0]
jbd = [0,d2,-d2]
jbu = [0,d2,d2]
jfd = [0,-d2,-d2]
jfu = [0,-d2,d2]
jbd2 = [0,d1,-d1]
jbu2 = [0,d1,d1]
jfd2 = [0,-d1,-d1]
jfu2 = [0,-d1,d1]
jlb2 = [-d1,d1,0]
jrf2 = [d1,-d1,0]
jlu3 = [-d3,d3,d3] # special for roof
jlb3 = [-d2,d1*1.1,0] # special roof
jrbd3 = [d3,d3,-d3] # special for roof
jlfd3 = [-d3,-d3,-d3] # special for roof
jrfd3 = [d3,-d3,-d3] # special for roof
jlbd3 = [-d3,d3,-d3] # special for roof
jrbu3 = [d3,d3,d3] # special for roof
jlbu3 = [-d3,d3,d3] # special for roof
jrfu3 = [d3,-d3,d3] # special for roof
jlfu3 = [-d3,-d3,d3] # special for roof
jrfd3 = [d3,-d3,-d3] # special for roof
jlfd3 = [-d3,-d3,-d3] # special for roof
jrbd3 = [d3,d3,-d3] # special
jlbd3 = [-d3,d3,-d3] # special
jlfu = [-d1,-d3,d2] # special for airbox to bonnet
jrfu = [d1,-d3,d2] # special for airbox to bonnet
jdr = [d2,0,-d2]
jdl = [-d2,0,-d2]
jdl3 = [-d3,0,-d3]
jur = [d2,0,d2]
jul = [-d2,0,d2]
jur2 = [d1,0,d1]
jul2 = [-d1,0,d1]
jur3 = [d2*1.1,0,d2*1.1] # special for front
jul3 = [-d2*1.1,0,d2*1.1] # special for front
jur4 = [d1*1.1,0,d1] # special for just behind front wheel floor
jul4 = [-d1*1.1,0,d1] # special for just behind front wheel floor
jbul = [-d2,d2,d2]
jbur = [d2,d2,d2]
jbdl = [-d2,d2,-d2]
jbdr = [d2,d2,-d2]
jfdr = [d2,-d2,-d2]
jfdl = [-d2,-d2,-d2]
jfur = [d2,-d2,d2]
jful = [-d2,-d2,d2]
jfur2 = [d2*1.1,-d1,d2]
jdl2 = [-d2*1.1,0,-d1] # special for diagonal on front bonnet to floor
jfdl3 = [-d1*0.55,-d2,-d2*1.1] # special for right-front bonnet to floor
jfdr3 = [d1*0.55,-d2,-d2*1.1] # special for right-front bonnet to floor
jfdl2 = [-d2,-d1,-d2]
jfdr2 = [d2,-d1,-d2]
jbdr2 = [d2,d1,-d2]
jbdl2 = [-d2,d1,-d2]
jfdl0 = [-d1/2,-d2/2,-d2/2] # special for bonnet front scoop
jfdl0 = [-d1/2,-d2/2,-d2/2] # special for bonnet front scoop
jfdr0 = [d1/2,-d2/2,-d2/2] # special for bonnet front scoop
jbul0 = [-d2/2,d2/2,d1/2] # special for bonnet front scoop
jbur0 = [d2/2,d2/2,d1/2] # special for bonnet front scoop
jleft0 = [-d2,0,0]
jright0 = [d2,0,0]
jleft3 = [-d1/2,0,0]
jright3 = [d1/2,0,0]
jdown0 = [0,0,-d2]
jup0 = [0,0,d2]
jfwd0 = [0,-d2*1.1,0,0]
jback0 = [0,d2*1.1,0]
jlb0 = [-d1/2,d1/2,0]
jrf0 = [d1/2,-d1/2,0]
jlf0 = [-d1/2,-d1/2,0]
jrb0 = [d1/2,d1/2,0]
jbu0 = [0,d1/2,d1/2]
jlbrot = rotate_x(jlb, -25) # really special back-box to rear
jrbrot = rotate_x(jrb, -25) # really special back-box to rear
jrbrot2 = rotate_x(rotate_y(jrb, 20), 70) # really special dashboard
jlbrot2 = rotate_x(rotate_y(jlb, -20), 70) # really special dashboard
jrbrot3 = rotate_z(jrb, -5) # really special airbox-to-back-box
jlbrot3 = rotate_z(jlb, 5) # really special airbox-to-back-box
jlfrot3 = rotate_z(jlf, -5) # really special airbox-to-back-box
jfrotl = rotate_z(jfwd, -25) # really special airbox-to-back-box
jfrotr = rotate_z(jfwd, 25) # really special airbox-to-back-box
jfdrot = rotate_x(jfd, 5) # really special back-box
jrfrot3 = rotate_z(jrf, 5) # really special back-box
jlfrot3 = rotate_z(jlf, -5) # really special back-box
jurrot = rotate_y(jur, -5) # really special back-box
jrrot = rotate_y(rotate_z(jright, -5), 5) # really special back-box
jlrot = rotate_y(rotate_z(jleft, 5), -5) # really special back-box
jrrot2 = rotate_y(jright, -5) # special back-box
jlrot2 = rotate_y(jleft, 5) # special back-box
jrbdrotz = rotate_z(jrbd3, 5) # really special back-box
jlbdrotz = rotate_z(jlbd3, -5) # really special back-box
jrfrotz2 = rotate_y(rotate_z(jrf, -5), -12) # really special rear section
jlfrotz2 = rotate_y(rotate_z(jlf, 5), 12) # really special rear section
jrfrot4 = rotate_z(jrf, -5) # really special bumper
jlfrot4 = rotate_z(jlf, 5) # really special bumper
jrbdrotz2 = rotate_z(jrbd3, 10) # really special bumper
jlbdrotz2 = rotate_z(jlbd3, -10) # really special bumper
jdrrot2 = rotate_y(jdr, 5) # really special rear section
jdlrot2 = rotate_y(jdl, -5) # really special rear section
jfurotz1 = rotate_z(jfu, 10) # really special rear section
jfurotz2 = rotate_z(jfu, -10) # really special rear section
jfrotu = rotate_x(jfwd, 10) # really special door sill

# longer joint-offset
LONGOFFS = 0.0

#auto-create joint offset
AUTO=1

#### TODO move to separate location
# wheel

def wheel():
    u = union()
    u.add(cylinder(h=wtire, r=wheel_radius, segments=40,center=True))
    u -= cylinder(h=wtire+0.1, r=wheel_inner, segments=20,center=True)
    r = rotate([90, 0, 90])
    r.add(u)
    return r


def auto_offset(p1, p2, diam, xangle=0.0, yangle=0.0, zangle=0.0):
    o = [0.0]*3
    pd = pdiff(p1, p2)
    pd = rotate_y(pd, -yangle)
    pd = rotate_x(pd, -xangle)
    l = pdist(p1, p2)
    for i in range(3):
        pd[i] /= l
        if abs(pd[i]) > 0.3:
            o[i] = math.copysign(1.0, pd[i])
    l = pdist([0,0,0], o)
    for i in range(3):
        o[i] /= l
    o = rotate_x(o, xangle)
    o = rotate_y(o, yangle)
    o = rotate_z(o, zangle)
    return o


def _joint(name, idx, pieces, offsetinset=0):
    u = union()
    jointoffs = set()

    # now add "interlock" pieces
    for p1, p2, diameter, offset1, offset2 in pieces:
        # create 5cm (0.005) "pipe" piece to stick into joint
        dmul = diameter*0.6
        _po1 = pmul(offset1, dmul) if offset1 else [0,0,0]
        _po2 = pmul(offset2, dmul) if offset2 else [0,0,0]
        jointoffs.add(tuple(_po1)) # add joint-point to set
        po1 = padd(p1, _po1)
        po2 = padd(p2, _po2)
        l = pdist(po1, po2)
        # pipe will sit inside this, an inset is removed later
        newpos = pavg(po1, po2, 0.005/l)
        newpos = pdiff(newpos, po2)
        p = create_pipe([0,0,0], newpos, diameter+0.004, offs=0.0)
        p = translate(_po1).add(p)
        u.add(p)

    radius = 0.008 / 2.0
    for pos in jointoffs:
        # TODO: bridge-joins between spheres (O(N^2))
        s = sphere(r=radius, segments=16)
        s = translate(list(pos)).add(s)
        u.add(s)
        if len(jointoffs) > 4: # shorten the list, by closest
            d = []
            for pos2 in jointoffs:
                if pos2 == pos:
                    continue
                l = pdist(pos, pos2)
                d.append((l, pos2))
            d.sort()
            jshorter = set()
            for (l, pos2) in d[:4]:
                jshorter.add(pos2)
        else:
            jshorter = jointoffs
        for pos2 in jshorter:
            if pos2 == pos:
                continue
            p = create_pipe(pos, pos2, 0.004, 0.005)
            u.add(p)

    # now remove "inset" pieces
    v = union()
    for p1, p2, diameter, offset1, offset2 in pieces:
        # create 5cm (0.005) "pipe" piece to stick into joint
        dmul = diameter*0.6
        _po1 = pmul(offset1, dmul) if offset1 else [0,0,0]
        _po2 = pmul(offset2, dmul) if offset2 else [0,0,0]
        jointoffs.add(tuple(_po1)) # add joint-point to set
        po1 = padd(p1, _po1)
        po2 = padd(p2, _po2)
        l = pdist(po1, po2)
        # pipe will sit inside this, an inset is removed later
        newpos = pavg(po1, po2, 0.02/l)
        newpos = pdiff(newpos, po2)
        p = create_pipe([0,0,0], newpos, diameter-0.0001, offs=0.0005)
        p = translate(_po1).add(p)
        v.add(p)

    # and outer (messy) pieces
    w = union()
    for p1, p2, diameter, offset1, offset2 in pieces:
        # create 5cm (0.005) "pipe" piece to stick into joint
        dmul = diameter*0.5999
        _po1 = pmul(offset1, dmul) if offset1 else [0,0,0]
        _po2 = pmul(offset2, dmul) if offset2 else [0,0,0]
        jointoffs.add(tuple(_po1)) # add joint-point to set
        po1 = padd(p1, _po1)
        po2 = padd(p2, _po2)
        l = pdist(po1, po2)
        # pipe will sit inside this, an inset is removed later
        newpos = pavg(po1, po2, 0.03/l)
        newpos = pdiff(newpos, po2)
        p = create_pipe([0,0,0], newpos, diameter+0.0041, offs=0.0100)
        p = translate(_po1).add(p)
        w.add(p)

    #d = difference()
    d = union()
    d.add(u)
    d.add(v)
    d.add(w)

    return d
    name = "joint_%s_%d" % (name, idx)
    openscad_to_stl(d, name)

    return import_stl(name)


def _joint(name, idx, pieces, offsetinset=0):
    u = union()
    jointoffs = set()

    # calculate "interlock" pieces
    for p1, p2, diameter, offset1, offset2 in pieces:
        # create 5cm (0.005) "pipe" piece to stick into joint
        dmul = diameter*0.4
        _po1 = pmul(offset1, dmul) if offset1 else [0,0,0]
        _po2 = pmul(offset2, dmul) if offset2 else [0,0,0]
        jointoffs.add(tuple(_po1)) # add joint-point to set
        po1 = padd(p1, _po1)
        po2 = padd(p2, _po2)
        l = pdist(po1, po2)
        # pipe will sit inside this, an inset is removed later
        newpos = pavg(po1, po2, 0.008/l)
        newpos = pdiff(newpos, po2)
        p = create_pipe([0,0,0], newpos, diameter+0.004, offs=0.0)
        p = translate(_po1).add(p)
        u.add(p)

    radius = 0.025 / 2.0
    s = sphere(r=radius, segments=16)
    u.add(s)

    # now remove "inset" pieces
    v = union()
    for p1, p2, diameter, offset1, offset2 in pieces:
        # create 5cm (0.005) "pipe" piece to stick into joint
        dmul = diameter*0.6
        _po1 = pmul(offset1, dmul) if offset1 else [0,0,0]
        _po2 = pmul(offset2, dmul) if offset2 else [0,0,0]
        jointoffs.add(tuple(_po1)) # add joint-point to set
        po1 = padd(p1, _po1)
        po2 = padd(p2, _po2)
        l = pdist(po1, po2)
        # pipe will sit inside this, an inset is removed later
        newpos = pavg(po1, po2, 0.02/l)
        newpos = pdiff(newpos, po2)
        p = create_pipe([0,0,0], newpos, diameter-0.0001, offs=0.0005)
        p = translate(_po1).add(p)
        v.add(p)

    # and outer (messy) pieces
    w = union()
    for p1, p2, diameter, offset1, offset2 in pieces:
        # create 5cm (0.005) "pipe" piece to stick into joint
        dmul = diameter*0.5999
        _po1 = pmul(offset1, dmul) if offset1 else [0,0,0]
        _po2 = pmul(offset2, dmul) if offset2 else [0,0,0]
        jointoffs.add(tuple(_po1)) # add joint-point to set
        po1 = padd(p1, _po1)
        po2 = padd(p2, _po2)
        l = pdist(po1, po2)
        # pipe will sit inside this, an inset is removed later
        newpos = pavg(po1, po2, 0.03/l)
        newpos = pdiff(newpos, po2)
        p = create_pipe([0,0,0], newpos, diameter+0.0041, offs=0.0100)
        p = translate(_po1).add(p)
        v.add(p)

    #d = difference()
    d = union()
    d.add(u)
    d.add(v)

    return d


def joint(name, idx, pieces, offsinset=0.0):
    u = union()
    jointoffs = set()
    t = union()

    # calculate "interlock" pieces
    for p1, p2, diameter, offset1, offset2 in pieces:
        # create 5cm (0.005) "pipe" piece to stick into joint
        dmul = diameter
        _po1 = pmul(offset1, dmul*1.0) if offset1 else [0,0,0]
        _po2 = pmul(offset2, dmul) if offset2 else [0,0,0]
        jointoffs.add(tuple(_po1)) # add joint-point to set
        po1 = padd(p1, _po1)
        po2 = padd(p2, _po2)
        l = pdist(po1, po2)
        # pipe will sit inside this, an inset is removed later
        if diameter == dthin: # inset
            newpos = pavg(po1, po2, 0.0093/l)
        else:                 # larger is not inset
            newpos = pavg(po1, po2, 0.0085/l)
        newpos = pdiff(newpos, po2)
        #if diameter == dthick:
        #    offsinset = 0.03
        offsinset = -0.000
        diameter_mod = diameter + 0.0015
        if diameter == dthin:
            diameter_mod += 0.0060
        p = create_pipe([0,0,0], newpos, diameter_mod, offs=offsinset,
                        diameter2=diameter_mod*0.30)
        p = translate(_po1).add(p)
        u.add(p)
        p = create_pipe(_po1, [0,0,0], offs=0.001,
                diameter=0.009, diameter2=0.009)
        u.add(p)
        # create piece to stick pipe against
        dmul = diameter
        print ("dmul", dmul)
        _po1 = pmul(offset1, dmul*1.0) if offset1 else [0,0,0]
        _po2 = pmul(offset2, dmul) if offset2 else [0,0,0]
        jointoffs.add(tuple(_po1)) # add joint-point to set
        po1 = padd(p1, _po1)
        po2 = padd(p2, _po2)
        l = pdist(po1, po2)
        # pipe will sit inside this, an inset is removed later
        if diameter == dthin:
            newpos = pavg(po1, po2, 0.050/l)
        else:
            newpos = pavg(po1, po2, 0.020/l)
        newpos = pdiff(newpos, po2)
        #if diameter == dthick:
        #    offsinset = 0.03
        if diameter == dthin:
            offsinset = 0.0165
            diameter_mod = diameter + 0.0005
        else:
            offsinset = 0.015
            diameter_mod = diameter - 0.0027
        p = create_pipe([0,0,0], newpos, diameter_mod, offs=offsinset,
                        diameter2=diameter_mod)
        p = translate(_po1).add(p)
        if diameter == dthin: # remove smaller pipes
            t.add(p)
        else:
            u.add(p)

    radius = 0.010 / 2.0
    for pos in jointoffs:
        # TODO: bridge-joins between spheres (O(N^2))
        s = sphere(r=radius, segments=16)
        s = translate(list(pos)).add(s)
        u.add(s)
    radius = 0.015 / 2.0
    s = sphere(r=radius, segments=16)
    u.add(s)

    d = difference()
    d.add(u)
    d.add(t)

    # hack-way to identify nodes: remove them, re-run... sigh
    #if name == 'Roof' and (idx in [0, 1, 2, 3, 4, 5]):
       #return union()
    #if name == 'Floor' and (idx in [0, 12]):
    #   return union()
    #if name == 'Airbox' and (idx in [8,11]):
    #   return union()
    #if name == 'BackBox' and (idx in [8, 9, 10, 11]):
    #   return union()
    #if name == 'floor_support' and (idx in [10]):
    #   return union()
    #if name == 'rear_bumper' and (idx in [13]):
    #   return union()

    # print out all (some?) joints to STL
    #if name == 'Bonnet':
    #if name == 'Roof' and (idx in [24, 22, 23, 25]):
    #if name == 'Floor' and (idx in [3, 7, 11, 15]):
    #if name == 'Airbox' and (idx in [8,11]):
    #if name == 'Floor' and (idx in [0, 4, 8, 12]):
    #if name == 'Roof' and (idx in [0, 5]):
    #if False:
    #if name == 'Floor' and (idx in [0, 12]):
    #if name == 'Roof' and (idx in [5]):
    #if name == 'BackBox': # and (idx in [0]):
    #if name == 'Roof' and (idx in [5]):
    #if name == 'floor_support':
    #if name == 'floor_support' and (idx in [10]):

    # save scad joints - post-processing of roof ones is needed
    # to join them together: see roof_imports.scad. TODO: in pyopenscad
    s = union()
    sc = scale([1000,1000,1000])
    sc = sc.add(d)
    s.add(sc)
    fname = "joint_%s_%d" % (name, idx)
    if name == 'Roof' and (idx in [6,7,12,13]):
        openscad_save(s, fname)
    if False:
    #if name in ['rear_bumper'] and idx in [2, 5]:
        openscad_to_stl(s, fname)
    #if name in ['rear_bumper', 'RearSection']:
    #    openscad_to_stl(s, fname)

    return d
    name = "joint_%s_%d" % (name, idx)
    openscad_to_stl(d, name)

    return import_stl(name)


class Model:
    def __init__(self):
        self.t = union()
        self.groups = []
        self.joints = []
        self.pipes = []
        self.extras = union()
        self.plen = 0.0

    def add(self, group, color, ops=None):
        group.color = color
        group.ops = ops
        self.groups.append(group)
        self.joints += group.joints
        self.pipes += group.pipes

    def find_xyzangle_for_pipe(self, p1, p2, rot):
        xangle1 = 0
        yangle1 = 0
        zangle1 = 0
        xangle2 = 0
        yangle2 = 0
        zangle2 = 0
        if rot:
            pipes = []
            for pos, xangle, yangle, zangle in self.joints:
                #print len(self.joints), p1, p2, pos, xangle
                if pos == p1:
                    xangle1 = xangle
                    yangle1 = yangle
                    zangle1 = zangle
                if pos == p2:
                    xangle2 = xangle
                    yangle2 = yangle
                    zangle2 = zangle
        return -xangle1, -yangle1, -zangle1, xangle2, yangle2, zangle2

    def find_pipes_for_joint(self, search):
        pipes = []
        for grp in self.groups:
            any_pipes = grp.pipes
            for p1, p2, diameter, offset1, offset2, rot in any_pipes:
                angles = self.find_xyzangle_for_pipe(p1, p2, rot)
                fxangle1, fyangle1, fzangle1, \
                fxangle2, fyangle2, fzangle2 = angles
                if offset1 == AUTO:
                    offset1 = auto_offset(p1, p2, diameter,
                                          fxangle1, fyangle1, fzangle1)
                if offset2 == AUTO:
                    offset2 = auto_offset(p2, p1, diameter,
                                          fxangle2, fyangle2, fzangle2)
                if pclose(search, p1, 0.0001):
                    pipes.append((p1, p2, diameter, offset1, offset2))
                if pclose(search, p2, 0.0001):
                    pipes.append((p2, p1, diameter, offset2, offset1))
        return pipes


    def get_joint_idx(self, point):
        for grp in self.groups:
            for idx, joint in enumerate(grp.joints):
                name = grp.name or grp.__class__.__name__
                #print ("find", name, idx, point, joint)
                if point == joint[0]:
                    return idx, name
        return "None", -1

    def _get_group(self, grp):
        name = grp.name or grp.__class__.__name__
        color = grp.color
        ops = grp.ops
        self.t.add(color)
        w = union()
        print "_get_group", name
        for j in self.joints:
            print "    ", j
        for p in grp.pipes:
            print "pipe ", p
        for idx, (pos, _, _, _) in enumerate(grp.joints):
            connections = self.find_pipes_for_joint(pos)
            w.add(translate(pos).add(joint(name, idx, connections, grp.offsi)))
        i = 0

        f = open("%s.txt" % name.lower(), "w")
        for (p1, p2, diameter, offset1, offset2, rot) in grp.pipes:
            idx1, grp1name = self.get_joint_idx(p1)
            idx2, grp2name = self.get_joint_idx(p2)
            angles = self.find_xyzangle_for_pipe(p1, p2, rot)
            fxangle1, fyangle1, fzangle1, fxangle2, fyangle2, fzangle2 = angles
            if offset1 == AUTO:
                offset1 = auto_offset(p1, p2, diameter,
                                      fxangle1, fyangle1, fzangle1)
            if offset2 == AUTO:
                offset2 = auto_offset(p2, p1, diameter,
                                      fxangle2, fyangle2, fzangle2)
            self.plen += pdist(p1, p2)
            dmul = diameter
            po1 = padd(p1, pmul(offset1, dmul)) if offset1 else p1
            po2 = padd(p2, pmul(offset2, dmul)) if offset2 else p2
            p = create_pipe(po1, po2, diameter, offs=0.025-grp.offsi)
            length = p.length
            if i != -1 or name != 'rear_bumper':
                w.add(p)
            # print out pipe lengths
            ha, va, fa = hangle(po1, po2), vangle(po1, po2), fangle(po1, po2)
            msg = "\t%s %d: %s/%s-%s/%s: %.1f a %d %d %d t %.1f" % \
                                                 (name, i,
                                                 grp1name, str(idx1),
                                                 grp2name, str(idx2),
                                                 #pdist(po1, po2)*100,
                                                 length*100,
                                                 ha, va, fa, diameter*1000)
            print msg
            f.write("%s\n\n" % msg)
            i += 1
        # close output report, generate PDF
        f.close()
        os.system("./create_pdf.sh %s" % name.lower())

        for (pos, angle) in grp.wheels:
            wh = wheel()
            wh = rotate([0,0, angle]).add(wh)
            w.add(translate(pos).add(wh))
        if ops:
            for op in ops:
                w = op.add(w)
        color.add(w)

    def get_model(self):
        for g in self.groups:
            g.do_doorline_pipes()
        for g in self.groups:
            if g.visible:
                self._get_group(g)
        self.t.add(self.extras) # add these after, due to transparency
        return self.t

class Group:
    def __init__(self, name=None):
        self.name = name
        self.joints = []
        self.pipes = []
        self.doorlinepipes = []
        self.wheels = []
        self.plen = 0.0
        self.dflt_pipe_diam = dthick
        self.visible = True
        self.offsi = 0.008

    def add_doorline_pipe(self, dooridx1, dooridx2, p1, p2, diameter=None):
        if diameter is None:
            diameter = self.dflt_pipe_diam
        self.doorlinepipes.append((dooridx1, dooridx2, p1, p2, diameter))

    def do_doorline_pipes(self):
        for (idx1, idx2, p1, p2, diameter) in self.doorlinepipes:
            off1, off2 = calc_door_offsets(idx1, True)
            #off1, off2 = AUTO, AUTO
            rot = False
            self.pipes.append((p1, p2, diameter, off1, off2, rot))

    def add_pipe(self, p1, p2, diameter=None, off1=AUTO, off2=AUTO, rot=True):
        #off1=AUTO
        #off2=AUTO
        if diameter is None:
            diameter = self.dflt_pipe_diam
        self.pipes.append((p1, p2, diameter, off1, off2, rot))

    def add_joint(self, pos, xangle=0, yangle=0, zangle=0):
        self.joints.append((pos, xangle, yangle, zangle))

    def add_wheel(self, pos, angle):
        self.wheels.append((pos, angle))


def create_pipe(p1, p2, diameter=dthick, offs=0.02,
                        cutangle1=None, cutangle2=None, diameter2=None):
    if diameter2 is None:
        diameter2 = diameter
    pipe = ppipe(p1, p2, radius=diameter*0.5, offs=offs,
                         cutangle1=cutangle1, cutangle2=cutangle2,
                         radius2=diameter2*0.5)
    length = pipe.length

    angle = theta(p1, p2)
    angle2 = hangle(p1, p2)
    pipe = rotate([0 , angle , 0]).add(pipe)
    pipe = rotate([0 , 0 , angle2]).add(pipe)
    pipe = translate(pavg(p1, p2)).add(pipe)
    pipe.length = length

    return pipe

